#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <commctrl.h>
#include <richedit.h>
#include "AkelDLL.h"
#include "StackFunc.h"
#include "Resources\Resource.h"


//Include stack functions
#define StackInsertAfter
#define StackInsertBefore
#define StackInsertIndex
#define StackDelete
#define StackClear
#include "StackFunc.h"

//Defines
#define DLLA_MACROS_EXEC  1

#define STRID_HOTKEY_EXISTS     1
#define STRID_CANNOT_OPEN_FILE  2
#define STRID_ALREADY_EXIST     3
#define STRID_CONFIRM_DELETE    4
#define STRID_THREAD_BUSY       5
#define STRID_MACRO             6
#define STRID_HOTKEY            7
#define STRID_RECORD_LISTITEM   8
#define STRID_RECORD            9
#define STRID_STOP              10
#define STRID_SAVE              11
#define STRID_DELETE            12
#define STRID_ASSIGN            13
#define STRID_OK                14

#define OF_MAIN            0x1
#define OF_WINDOWRECT      0x2

#define LVI_MACRO_FILE          0
#define LVI_MACRO_HOTKEY        1

#define BUFFER_SIZE             1024
#define MACRO_TIMEOUT           5000

//DIALOGRESIZE
#define DRS_SIZE  0x1 //Resize control
#define DRS_MOVE  0x2 //Move control
#define DRS_X     0x4 //X value
#define DRS_Y     0x8 //Y value

#ifndef LVM_SETEXTENDEDLISTVIEWSTYLE
  #define LVM_SETEXTENDEDLISTVIEWSTYLE (LVM_FIRST + 54)
#endif
#ifndef LVS_EX_GRIDLINES
  #define LVS_EX_GRIDLINES 0x00000001
#endif
#ifndef LVS_EX_FULLROWSELECT
  #define LVS_EX_FULLROWSELECT 0x00000020
#endif

typedef struct {
  char *pMacro;
  int nRepeat;
  HANDLE hInitMutex;
} EXECMACROA;

typedef struct {
  wchar_t *wpMacro;
  int nRepeat;
  HANDLE hInitMutex;
} EXECMACROW;

typedef struct _KEYACT {
  BYTE bVk;
  DWORD dwFlags;
} KEYACT;

typedef struct _KEYSTRUCT {
  struct _KEYSTRUCT *next;
  struct _KEYSTRUCT *prev;
  KEYACT ka;
} KEYSTRUCT;

typedef struct {
  HWND *lpWnd;   //Control window
  DWORD dwType;  //See DRS_* defines
} DIALOGRESIZE;

//Functions prototypes
DWORD WINAPI ThreadProcA(LPVOID lpParameter);
DWORD WINAPI ThreadProcW(LPVOID lpParameter);
LRESULT CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK GetMsgProc(int code, WPARAM wParam, LPARAM lParam);
BOOL RegisterHotkeyA(char *szMacroName, WORD wHotkey);
BOOL RegisterHotkeyW(wchar_t *wszMacroName, WORD wHotkey);
BOOL CALLBACK HotkeyProcA(void *lpParameter);
BOOL CALLBACK HotkeyProcW(void *lpParameter);
DWORD WINAPI ExecThreadProcA(LPVOID lpParameter);
DWORD WINAPI ExecThreadProcW(LPVOID lpParameter);
KEYSTRUCT* StackHotkeyAdd(HSTACK *hStack, BYTE bVk, DWORD dwFlags);
void StackHotkeyPress(HSTACK *hStack, BOOL bToEnd);
BOOL StackHotkeyRead(HSTACK *hStack, HANDLE hFile);
BOOL StackHotkeySave(HSTACK *hStack, HANDLE hFile);
void StackHotkeyFree(HSTACK *hStack);
BOOL ReadMacroFileA(HSTACK *hKeyStack, char *pMacro);
BOOL ReadMacroFileW(HSTACK *hKeyStack, wchar_t *wpMacro);
BOOL SaveMacroFileA(HSTACK *hRecordStack, char *pMacro);
BOOL SaveMacroFileW(HSTACK *hRecordStack, wchar_t *wpMacro);
BOOL DeleteMacroFileA(char *pMacro);
BOOL DeleteMacroFileW(wchar_t *wpMacro);
void FillMacroListA(HWND hWnd);
void FillMacroListW(HWND hWnd);
void SaveMacroListA(HWND hWnd);
void SaveMacroListW(HWND hWnd);
void MacroRecord();
void MacroStop();
void MacroPlay(HSTACK *hStack, int nRepeat);
BOOL WaitIdle();
void ExecMacroA(char *pMacro, int nRepeat);
void ExecMacroW(wchar_t *wpMacro, int nRepeat);
LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
void GetHotkeyStringA(WORD wHotkey, char *szString);
void GetHotkeyStringW(WORD wHotkey, wchar_t *wszString);
int MoveListViewItemA(HWND hWnd, int nOldIndex, int nNewIndex);
int MoveListViewItemW(HWND hWnd, int nOldIndex, int nNewIndex);
int GetBaseNameA(char *pFile, char *szBaseName, int nBaseNameMaxLen);
int GetBaseNameW(wchar_t *wpFile, wchar_t *wszBaseName, int nBaseNameMaxLen);
BOOL FileExistsA(char *pFile);
BOOL FileExistsW(wchar_t *wpFile);
BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc);
BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcInit, RECT *rcCurrent, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
void ReadOptionsA();
void ReadOptionsW();
void SaveOptionsA(DWORD dwFlags);
void SaveOptionsW(DWORD dwFlags);
char* GetLangStringA(LANGID wLangID, int nStringID);
wchar_t* GetLangStringW(LANGID wLangID, int nStringID);
BOOL IsExtCallParamValid(LPARAM lParam, int nIndex);
int GetExtCallParam(LPARAM lParam, int nIndex);
void InitCommon(PLUGINDATA *pd);
void InitMain();
void UninitMain();

//Global variables
char szPluginName[MAX_PATH];
wchar_t wszPluginName[MAX_PATH];
HINSTANCE hInstanceDLL;
HWND hMainWnd;
BOOL bOldWindows;
BOOL bMDI;
HICON hMainIcon;
LANGID wLangSystem;
BOOL bInitCommon=FALSE;
BOOL bInitMain=FALSE;
WNDPROCDATA *NewMainProcData=NULL;
char szMutexInitName[]="AkelPad::Macros::MutexInit";
wchar_t wszMutexInitName[]=L"AkelPad::Macros::MutexInit";
char szMutexExecName[]="AkelPad::Macros::MutexExec";
wchar_t wszMutexExecName[]=L"AkelPad::Macros::MutexExec";
char szMacrosDir[MAX_PATH];
wchar_t wszMacrosDir[MAX_PATH];
char szBuffer[BUFFER_SIZE];
wchar_t wszBuffer[BUFFER_SIZE];
char szLastMacro[MAX_PATH]="";
wchar_t wszLastMacro[MAX_PATH]=L"";
HSTACK hRecordStack={0};
HSTACK hLastMacroStack={0};
HWND hWndMainDlg=NULL;
RECT rcMainInitDialog={0};
RECT rcMainCurrentDialog={0};
BOOL bMainDialogRectSave=FALSE;
int nColumnWidth1=160;
int nColumnWidth2=109;
HHOOK hHook=NULL;
HANDLE hDialogThread=NULL;
HANDLE hExecThread=NULL;
DWORD dwMainProcessId=0;
DWORD dwMainThreadId=0;
DWORD dwDialogThreadId=0;
DWORD dwExecThreadId=0;
BOOL bRecord=FALSE;
BOOL bPlay=FALSE;


//Identification
extern "C" void __declspec(dllexport) DllAkelPadID(PLUGINVERSION *pv)
{
  pv->dwAkelDllVersion=AKELDLL;
  pv->dwExeMinVersion3x=MAKE_IDENTIFIER(3, 7, 4, 0);
  pv->dwExeMinVersion4x=MAKE_IDENTIFIER(4, 4, 0, 0);
  pv->pPluginName="Macros";
}

//Plugin extern function
extern "C" void __declspec(dllexport) Main(PLUGINDATA *pd)
{
  if (pd->lpbAutoLoad)
  {
    *pd->lpbAutoLoad=TRUE;
    return;
  }

  if (!bInitCommon) InitCommon(pd);
  if (!bInitMain) InitMain();

  if (pd->lParam)
  {
    if (IsExtCallParamValid(pd->lParam, 1))
    {
      if (bOldWindows)
      {
        if (GetExtCallParam(pd->lParam, 1) == DLLA_MACROS_EXEC)
        {
          char *pMacro=NULL;
          int nRepeat=1;

          if (IsExtCallParamValid(pd->lParam, 2))
            pMacro=(char *)GetExtCallParam(pd->lParam, 2);
          if (IsExtCallParamValid(pd->lParam, 3))
            nRepeat=GetExtCallParam(pd->lParam, 3);

          if (pMacro)
          {
            ExecMacroA(pMacro, nRepeat);
          }
        }
      }
      else
      {
        if (GetExtCallParam(pd->lParam, 1) == DLLA_MACROS_EXEC)
        {
          wchar_t *wpMacro=NULL;
          int nRepeat=1;

          if (IsExtCallParamValid(pd->lParam, 2))
            wpMacro=(wchar_t *)GetExtCallParam(pd->lParam, 2);
          if (IsExtCallParamValid(pd->lParam, 3))
            nRepeat=GetExtCallParam(pd->lParam, 3);

          if (wpMacro)
          {
            ExecMacroW(wpMacro, nRepeat);
          }
        }
      }
    }
  }
  else
  {
    if (!pd->bOnStart)
    {
      if (!hDialogThread)
      {
        if (bOldWindows)
          hDialogThread=CreateThread(NULL, 0, ThreadProcA, NULL, 0, &dwDialogThreadId);
        else
          hDialogThread=CreateThread(NULL, 0, ThreadProcW, NULL, 0, &dwDialogThreadId);
      }
    }
  }

  //Stay in memory, and show as active
  pd->nUnload=UD_NONUNLOAD_ACTIVE;
}

DWORD WINAPI ThreadProcA(LPVOID lpParameter)
{
  MSG msg;
  BOOL bMsgStatus;

  hWndMainDlg=CreateDialogA(hInstanceDLL, MAKEINTRESOURCEA(IDD_MACROS), hMainWnd, (DLGPROC)MainDlgProcA);

  if (hWndMainDlg)
  {
    while ((bMsgStatus=GetMessage(&msg, NULL, 0, 0)) && bMsgStatus != -1)
    {
      if (!IsDialogMessage(hWndMainDlg, &msg))
      {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
      }
    }
  }

  return 0;
}

DWORD WINAPI ThreadProcW(LPVOID lpParameter)
{
  MSG msg;
  BOOL bMsgStatus;

  hWndMainDlg=CreateDialogW(hInstanceDLL, MAKEINTRESOURCEW(IDD_MACROS), hMainWnd, (DLGPROC)MainDlgProcW);

  if (hWndMainDlg)
  {
    while ((bMsgStatus=GetMessageW(&msg, NULL, 0, 0)) && bMsgStatus != -1)
    {
      if (!IsDialogMessageW(hWndMainDlg, &msg))
      {
        TranslateMessage(&msg);
        DispatchMessageW(&msg);
      }
    }
  }

  return 0;
}

LRESULT CALLBACK MainDlgProcA(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndList;
  static HWND hWndPlayButton;
  static HWND hWndPlayX;
  static HWND hWndRepeat;
  static HWND hWndRepeatSpin;
  static HWND hWndRecordButton;
  static HWND hWndName;
  static HWND hWndSaveButton;
  static HWND hWndDeleteButton;
  static HWND hWndHotkey;
  static HWND hWndAssignButton;
  static HWND hWndOKButton;
  static HICON hIconPlay;
  static int nSelItem=-1;
  static BOOL bListChanged=FALSE;
  static DIALOGRESIZE drs[]={{&hWndList,         DRS_SIZE|DRS_X|DRS_Y},
                             {&hWndPlayButton,   DRS_MOVE|DRS_X},
                             {&hWndPlayX,        DRS_MOVE|DRS_X},
                             {&hWndRepeat,       DRS_MOVE|DRS_X},
                             {&hWndRepeatSpin,   DRS_MOVE|DRS_X},
                             {&hWndRecordButton, DRS_MOVE|DRS_X},
                             {&hWndName,         DRS_MOVE|DRS_X},
                             {&hWndSaveButton,   DRS_MOVE|DRS_X},
                             {&hWndDeleteButton, DRS_MOVE|DRS_X},
                             {&hWndHotkey,       DRS_MOVE|DRS_X},
                             {&hWndAssignButton, DRS_MOVE|DRS_X},
                             {&hWndOKButton,     DRS_MOVE|DRS_X|DRS_Y},
                             {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    LVCOLUMNA lvc;

    SendMessage(hDlg, WM_SETICON, ICON_BIG, (LPARAM)hMainIcon);
    hIconPlay=(HICON)LoadImageA(hInstanceDLL, MAKEINTRESOURCEA(IDI_ICON_PLAY), IMAGE_ICON, 0, 0, 0);
    hWndList=GetDlgItem(hDlg, IDC_MACROS_LIST);
    hWndPlayButton=GetDlgItem(hDlg, IDC_PLAY);
    hWndPlayX=GetDlgItem(hDlg, IDC_PLAY_X);
    hWndRepeat=GetDlgItem(hDlg, IDC_REPEAT);
    hWndRepeatSpin=GetDlgItem(hDlg, IDC_REPEAT_SPIN);
    hWndRecordButton=GetDlgItem(hDlg, IDC_RECORD);
    hWndName=GetDlgItem(hDlg, IDC_NAME);
    hWndSaveButton=GetDlgItem(hDlg, IDC_SAVE);
    hWndDeleteButton=GetDlgItem(hDlg, IDC_DELETE);
    hWndHotkey=GetDlgItem(hDlg, IDC_HOTKEY);
    hWndAssignButton=GetDlgItem(hDlg, IDC_ASSIGN);
    hWndOKButton=GetDlgItem(hDlg, IDOK);

    SendMessage(hWndPlayButton, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIconPlay);
    SendMessage(hWndRepeatSpin, UDM_SETBUDDY, (WPARAM)hWndRepeat, 0);
    SendMessage(hWndRepeatSpin, UDM_SETRANGE, 0, MAKELONG(999, 0));
    SetDlgItemInt(hDlg, IDC_REPEAT, 1, FALSE);
    SendMessage(hWndName, EM_LIMITTEXT, MAX_PATH, 0);
    EnableWindow(hWndPlayButton, FALSE);
    EnableWindow(hWndRepeat, FALSE);
    EnableWindow(hWndRepeatSpin, FALSE);
    EnableWindow(hWndRecordButton, FALSE);
    EnableWindow(hWndName, FALSE);
    EnableWindow(hWndSaveButton, FALSE);
    EnableWindow(hWndDeleteButton, FALSE);
    EnableWindow(hWndHotkey, FALSE);
    EnableWindow(hWndAssignButton, FALSE);
    SendMessage(hWndList, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES);

    SetDlgItemTextA(hDlg, IDC_RECORD, GetLangStringA(wLangSystem, STRID_RECORD));
    SetDlgItemTextA(hDlg, IDC_SAVE, GetLangStringA(wLangSystem, STRID_SAVE));
    SetDlgItemTextA(hDlg, IDC_DELETE, GetLangStringA(wLangSystem, STRID_DELETE));
    SetDlgItemTextA(hDlg, IDC_ASSIGN, GetLangStringA(wLangSystem, STRID_ASSIGN));
    SetDlgItemTextA(hDlg, IDOK, GetLangStringA(wLangSystem, STRID_OK));

    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndHotkey, 0);

    //Columns
    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringA(wLangSystem, STRID_MACRO);
    lvc.cx=nColumnWidth1;
    lvc.iSubItem=LVI_MACRO_FILE;
    SendMessage(hWndList, LVM_INSERTCOLUMNA, LVI_MACRO_FILE, (LPARAM)&lvc);

    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringA(wLangSystem, STRID_HOTKEY);
    lvc.cx=nColumnWidth2;
    lvc.iSubItem=LVI_MACRO_HOTKEY;
    SendMessage(hWndList, LVM_INSERTCOLUMNA, LVI_MACRO_HOTKEY, (LPARAM)&lvc);

    FillMacroListA(hWndList);
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (wParam == IDC_MACROS_LIST)
    {
      if (((NMLISTVIEW *)lParam)->hdr.code == LVN_ITEMCHANGED)
      {
        if (((NMLISTVIEW *)lParam)->uNewState & LVIS_STATEIMAGEMASK)
        {
          BOOL bNewState;
          BOOL bOldState;

          bNewState=((((NMLISTVIEW *)lParam)->uNewState & LVIS_STATEIMAGEMASK) >> 12) - 1;
          bOldState=((((NMLISTVIEW *)lParam)->uOldState & LVIS_STATEIMAGEMASK) >> 12) - 1;

          if (bNewState >=0 && bOldState >=0 && bNewState != bOldState)
            bListChanged=TRUE;
        }
        if (((NMLISTVIEW *)lParam)->uNewState & LVIS_SELECTED)
        {
          SendMessage(hWndHotkey, HKM_SETHOTKEY, ((NMLISTVIEW *)lParam)->lParam, 0);
          nSelItem=((NMLISTVIEW *)lParam)->iItem;

          if (nSelItem > 0)
          {
            EnableWindow(hWndPlayButton, TRUE);
            EnableWindow(hWndRepeat, TRUE);
            EnableWindow(hWndRepeatSpin, TRUE);
            EnableWindow(hWndRecordButton, FALSE);
            EnableWindow(hWndName, FALSE);
            EnableWindow(hWndSaveButton, FALSE);
            EnableWindow(hWndDeleteButton, TRUE);
            EnableWindow(hWndHotkey, TRUE);
            EnableWindow(hWndAssignButton, TRUE);
          }
          else
          {
            if (hRecordStack.last)
            {
              EnableWindow(hWndPlayButton, TRUE);
              EnableWindow(hWndRepeat, TRUE);
              EnableWindow(hWndRepeatSpin, TRUE);
              EnableWindow(hWndName, TRUE);
              if (GetWindowTextLength(hWndName))
                EnableWindow(hWndSaveButton, TRUE);
              else
                EnableWindow(hWndSaveButton, FALSE);
            }
            else
            {
              EnableWindow(hWndPlayButton, FALSE);
              EnableWindow(hWndRepeat, FALSE);
              EnableWindow(hWndRepeatSpin, FALSE);
              EnableWindow(hWndName, FALSE);
              EnableWindow(hWndSaveButton, FALSE);
            }
            EnableWindow(hWndRecordButton, TRUE);
            EnableWindow(hWndDeleteButton, FALSE);
            EnableWindow(hWndHotkey, FALSE);
            EnableWindow(hWndAssignButton, FALSE);
          }
        }
        if (((NMLISTVIEW *)lParam)->uOldState & LVIS_SELECTED)
        {
          SendMessage(hWndHotkey, HKM_SETHOTKEY, 0, 0);
          nSelItem=-1;

          EnableWindow(hWndPlayButton, FALSE);
          EnableWindow(hWndRepeat, FALSE);
          EnableWindow(hWndRepeatSpin, FALSE);
          EnableWindow(hWndRecordButton, FALSE);
          EnableWindow(hWndName, FALSE);
          EnableWindow(hWndSaveButton, FALSE);
          EnableWindow(hWndDeleteButton, FALSE);
          EnableWindow(hWndHotkey, FALSE);
          EnableWindow(hWndAssignButton, FALSE);
        }
      }
      else if ((int)((NMHDR *)lParam)->code == NM_DBLCLK) PostMessage(hDlg, WM_COMMAND, IDC_PLAY, 0);
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_PLAY)
    {
      char szName[MAX_PATH];
      LVITEMA lvi;
      int nRepeat;

      if (nSelItem == 0)
      {
        EnableWindow(hWndPlayButton, FALSE);
        SetFocus(hMainWnd);
        nRepeat=GetDlgItemInt(hDlg, IDC_REPEAT, NULL, FALSE);
        MacroPlay(&hRecordStack, nRepeat);
        EnableWindow(hWndPlayButton, TRUE);
      }
      else
      {
        //Get macro name
        lvi.mask=LVIF_TEXT;
        lvi.pszText=szName;
        lvi.cchTextMax=MAX_PATH;
        lvi.iItem=nSelItem;
        lvi.iSubItem=LVI_MACRO_FILE;
        SendMessage(hWndList, LVM_GETITEMA, 0, (LPARAM)&lvi);

        if (lstrcmpiA(szName, szLastMacro))
        {
          StackHotkeyFree(&hLastMacroStack);
          szLastMacro[0]='\0';

          if (ReadMacroFileA(&hLastMacroStack, szName))
            lstrcpynA(szLastMacro, szName, MAX_PATH);
        }
        EnableWindow(hWndPlayButton, FALSE);
        SetFocus(hMainWnd);
        nRepeat=GetDlgItemInt(hDlg, IDC_REPEAT, NULL, FALSE);
        MacroPlay(&hLastMacroStack, nRepeat);
        EnableWindow(hWndPlayButton, TRUE);
      }
    }
    else if (LOWORD(wParam) == IDC_RECORD)
    {
      if (!bRecord)
      {
        EnableWindow(hWndPlayButton, FALSE);
        EnableWindow(hWndRepeat, FALSE);
        EnableWindow(hWndRepeatSpin, FALSE);
        EnableWindow(hWndName, FALSE);
        EnableWindow(hWndSaveButton, FALSE);
        EnableWindow(hWndList, FALSE);

        StackHotkeyFree(&hRecordStack);
        SetWindowTextA(hWndRecordButton, GetLangStringA(wLangSystem, STRID_STOP));
        SetFocus(hMainWnd);
        MacroRecord();
        bRecord=TRUE;
      }
      else
      {
        bRecord=FALSE;
        MacroStop();
        SetWindowTextA(hWndRecordButton, GetLangStringA(wLangSystem, STRID_RECORD));

        EnableWindow(hWndList, TRUE);
        if (hRecordStack.last)
        {
          EnableWindow(hWndPlayButton, TRUE);
          EnableWindow(hWndRepeat, TRUE);
          EnableWindow(hWndRepeatSpin, TRUE);
          EnableWindow(hWndName, TRUE);
          if (GetWindowTextLengthW(hWndName))
            EnableWindow(hWndSaveButton, TRUE);
          else
            EnableWindow(hWndSaveButton, FALSE);
        }
      }
    }
    else if (LOWORD(wParam) == IDC_NAME)
    {
      if (GetWindowTextLength(hWndName))
        EnableWindow(hWndSaveButton, TRUE);
      else
        EnableWindow(hWndSaveButton, FALSE);
    }
    else if (LOWORD(wParam) == IDC_SAVE)
    {
      char szName[MAX_PATH];
      LVITEMA lvi;
      int nIndex;

      if (GetWindowTextA(hWndName, szName, MAX_PATH))
      {
        wsprintfA(szBuffer, "%s\\%s.macro", szMacrosDir, szName);

        if (!FileExistsA(szBuffer))
        {
          if (SaveMacroFileA(&hRecordStack, szName))
          {
            lstrcpynA(szLastMacro, szName, MAX_PATH);
            memcpy(&hLastMacroStack, &hRecordStack, sizeof(HSTACK));
            hRecordStack.first=0;
            hRecordStack.last=0;

            lvi.mask=LVIF_TEXT;
            lvi.pszText=szName;
            lvi.iItem=SendMessage(hWndList, LVM_GETITEMCOUNT, 0, 0);
            lvi.iSubItem=LVI_MACRO_FILE;
            nIndex=SendMessage(hWndList, LVM_INSERTITEMA, 0, (LPARAM)&lvi);

            if (!nIndex) nIndex=MoveListViewItemA(hWndList, nIndex, nIndex + 1);

            SetFocus(hWndList);
            lvi.stateMask=LVIS_SELECTED|LVIS_FOCUSED;
            lvi.state=LVIS_SELECTED|LVIS_FOCUSED;
            SendMessage(hWndList, LVM_SETITEMSTATE, nIndex, (LPARAM)&lvi);
          }
          else MessageBoxA(hDlg, GetLangStringA(wLangSystem, STRID_CANNOT_OPEN_FILE), "Macros plugin", MB_OK|MB_ICONERROR);
        }
        else MessageBoxA(hDlg, GetLangStringA(wLangSystem, STRID_ALREADY_EXIST), "Macros plugin", MB_OK|MB_ICONEXCLAMATION);
      }
    }
    else if (LOWORD(wParam) == IDC_DELETE)
    {
      char szName[MAX_PATH];
      LVITEMA lvi;

      if (nSelItem > 0)
      {
        if (MessageBoxA(hDlg, GetLangStringA(wLangSystem, STRID_CONFIRM_DELETE), "Macros plugin", MB_YESNO|MB_ICONQUESTION) == IDYES)
        {
          //Get macro name
          lvi.mask=LVIF_TEXT;
          lvi.pszText=szName;
          lvi.cchTextMax=MAX_PATH;
          lvi.iItem=nSelItem;
          lvi.iSubItem=LVI_MACRO_FILE;
          SendMessage(hWndList, LVM_GETITEMA, 0, (LPARAM)&lvi);

          if (DeleteMacroFileA(szName))
          {
            if (!lstrcmpiA(szName, szLastMacro))
            {
              StackHotkeyFree(&hLastMacroStack);
              szLastMacro[0]='\0';
            }
            SendMessage(hWndList, LVM_DELETEITEM, nSelItem, 0);
          }
        }
      }
    }
    else if (LOWORD(wParam) == IDC_ASSIGN)
    {
      LVITEMA lvi;
      WORD wHotkey;
      char szFile[MAX_PATH];
      char szHotkey[MAX_PATH];

      //Get macro name
      lvi.mask=LVIF_TEXT;
      lvi.pszText=szFile;
      lvi.cchTextMax=MAX_PATH;
      lvi.iItem=nSelItem;
      lvi.iSubItem=LVI_MACRO_FILE;
      SendMessage(hWndList, LVM_GETITEMA, 0, (LPARAM)&lvi);

      //Get macro hotkey
      wHotkey=(WORD)SendMessage(hWndHotkey, HKM_GETHOTKEY, 0, 0);

      //Register hotkey
      if (RegisterHotkeyA(szFile, wHotkey))
      {
        GetHotkeyStringA(wHotkey, szHotkey);

        lvi.mask=LVIF_PARAM;
        lvi.iItem=nSelItem;
        lvi.iSubItem=LVI_MACRO_FILE;
        lvi.lParam=wHotkey;
        SendMessage(hWndList, LVM_SETITEMA, 0, (LPARAM)&lvi);

        lvi.mask=LVIF_TEXT;
        lvi.pszText=szHotkey;
        lvi.iItem=nSelItem;
        lvi.iSubItem=LVI_MACRO_HOTKEY;
        SendMessage(hWndList, LVM_SETITEMA, 0, (LPARAM)&lvi);

        SendMessage(hWndHotkey, HKM_SETHOTKEY, wHotkey, 0);
        bListChanged=TRUE;
      }
      else SetFocus(hWndHotkey);
    }
    else if (LOWORD(wParam) == IDOK ||
             LOWORD(wParam) == IDCANCEL)
    {
      int nWidth;

      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_MACRO_FILE, 0);
      if (nWidth != nColumnWidth1)
      {
        nColumnWidth1=nWidth;
        bMainDialogRectSave=TRUE;
      }
      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_MACRO_HOTKEY, 0);
      if (nWidth != nColumnWidth2)
      {
        nColumnWidth2=nWidth;
        bMainDialogRectSave=TRUE;
      }

      if (bListChanged || LOWORD(wParam) == IDOK)
      {
        SaveMacroListA(hWndList);
        SaveOptionsA(OF_MAIN|OF_WINDOWRECT);
      }

      if (bRecord)
      {
        bRecord=FALSE;
        MacroStop();
      }
      DestroyIcon(hIconPlay);

      DestroyWindow(hWndMainDlg);
      hWndMainDlg=NULL;
      CloseHandle(hDialogThread);
      hDialogThread=NULL;
      ExitThread(0);
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }

  if (DialogResizeMessages(&drs[0], &rcMainInitDialog, &rcMainCurrentDialog, hDlg, uMsg, wParam, lParam))
    bMainDialogRectSave=TRUE;

  return FALSE;
}

LRESULT CALLBACK MainDlgProcW(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  static HWND hWndList;
  static HWND hWndPlayButton;
  static HWND hWndPlayX;
  static HWND hWndRepeat;
  static HWND hWndRepeatSpin;
  static HWND hWndRecordButton;
  static HWND hWndName;
  static HWND hWndSaveButton;
  static HWND hWndDeleteButton;
  static HWND hWndHotkey;
  static HWND hWndAssignButton;
  static HWND hWndOKButton;
  static HICON hIconPlay;
  static int nSelItem=-1;
  static BOOL bListChanged=FALSE;
  static DIALOGRESIZE drs[]={{&hWndList,         DRS_SIZE|DRS_X|DRS_Y},
                             {&hWndPlayButton,   DRS_MOVE|DRS_X},
                             {&hWndPlayX,        DRS_MOVE|DRS_X},
                             {&hWndRepeat,       DRS_MOVE|DRS_X},
                             {&hWndRepeatSpin,   DRS_MOVE|DRS_X},
                             {&hWndRecordButton, DRS_MOVE|DRS_X},
                             {&hWndName,         DRS_MOVE|DRS_X},
                             {&hWndSaveButton,   DRS_MOVE|DRS_X},
                             {&hWndDeleteButton, DRS_MOVE|DRS_X},
                             {&hWndHotkey,       DRS_MOVE|DRS_X},
                             {&hWndAssignButton, DRS_MOVE|DRS_X},
                             {&hWndOKButton,     DRS_MOVE|DRS_X|DRS_Y},
                             {0, 0}};

  if (uMsg == WM_INITDIALOG)
  {
    LVCOLUMNW lvc;

    SendMessage(hDlg, WM_SETICON, ICON_BIG, (LPARAM)hMainIcon);
    hIconPlay=(HICON)LoadImageA(hInstanceDLL, MAKEINTRESOURCEA(IDI_ICON_PLAY), IMAGE_ICON, 0, 0, 0);
    hWndList=GetDlgItem(hDlg, IDC_MACROS_LIST);
    hWndPlayButton=GetDlgItem(hDlg, IDC_PLAY);
    hWndPlayX=GetDlgItem(hDlg, IDC_PLAY_X);
    hWndRepeat=GetDlgItem(hDlg, IDC_REPEAT);
    hWndRepeatSpin=GetDlgItem(hDlg, IDC_REPEAT_SPIN);
    hWndRecordButton=GetDlgItem(hDlg, IDC_RECORD);
    hWndName=GetDlgItem(hDlg, IDC_NAME);
    hWndSaveButton=GetDlgItem(hDlg, IDC_SAVE);
    hWndDeleteButton=GetDlgItem(hDlg, IDC_DELETE);
    hWndHotkey=GetDlgItem(hDlg, IDC_HOTKEY);
    hWndAssignButton=GetDlgItem(hDlg, IDC_ASSIGN);
    hWndOKButton=GetDlgItem(hDlg, IDOK);

    SendMessage(hWndPlayButton, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIconPlay);
    SendMessage(hWndRepeatSpin, UDM_SETBUDDY, (WPARAM)hWndRepeat, 0);
    SendMessage(hWndRepeatSpin, UDM_SETRANGE, 0, MAKELONG(999, 0));
    SetDlgItemInt(hDlg, IDC_REPEAT, 1, FALSE);
    SendMessage(hWndName, EM_LIMITTEXT, MAX_PATH, 0);
    EnableWindow(hWndPlayButton, FALSE);
    EnableWindow(hWndRepeat, FALSE);
    EnableWindow(hWndRepeatSpin, FALSE);
    EnableWindow(hWndRecordButton, FALSE);
    EnableWindow(hWndName, FALSE);
    EnableWindow(hWndSaveButton, FALSE);
    EnableWindow(hWndDeleteButton, FALSE);
    EnableWindow(hWndHotkey, FALSE);
    EnableWindow(hWndAssignButton, FALSE);
    SendMessage(hWndList, LVM_SETEXTENDEDLISTVIEWSTYLE, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES, LVS_EX_FULLROWSELECT|LVS_EX_GRIDLINES);

    SetDlgItemTextW(hDlg, IDC_RECORD, GetLangStringW(wLangSystem, STRID_RECORD));
    SetDlgItemTextW(hDlg, IDC_SAVE, GetLangStringW(wLangSystem, STRID_SAVE));
    SetDlgItemTextW(hDlg, IDC_DELETE, GetLangStringW(wLangSystem, STRID_DELETE));
    SetDlgItemTextW(hDlg, IDC_ASSIGN, GetLangStringW(wLangSystem, STRID_ASSIGN));
    SetDlgItemTextW(hDlg, IDOK, GetLangStringW(wLangSystem, STRID_OK));

    SendMessage(hMainWnd, AKD_SETHOTKEYINPUT, (WPARAM)hWndHotkey, 0);

    //Columns
    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringW(wLangSystem, STRID_MACRO);
    lvc.cx=nColumnWidth1;
    lvc.iSubItem=LVI_MACRO_FILE;
    SendMessage(hWndList, LVM_INSERTCOLUMNW, LVI_MACRO_FILE, (LPARAM)&lvc);

    lvc.mask=LVCF_TEXT|LVCF_WIDTH|LVCF_SUBITEM;
    lvc.pszText=GetLangStringW(wLangSystem, STRID_HOTKEY);
    lvc.cx=nColumnWidth2;
    lvc.iSubItem=LVI_MACRO_HOTKEY;
    SendMessage(hWndList, LVM_INSERTCOLUMNW, LVI_MACRO_HOTKEY, (LPARAM)&lvc);

    FillMacroListW(hWndList);
  }
  else if (uMsg == WM_NOTIFY)
  {
    if (wParam == IDC_MACROS_LIST)
    {
      if (((NMLISTVIEW *)lParam)->hdr.code == LVN_ITEMCHANGED)
      {
        if (((NMLISTVIEW *)lParam)->uNewState & LVIS_STATEIMAGEMASK)
        {
          BOOL bNewState;
          BOOL bOldState;

          bNewState=((((NMLISTVIEW *)lParam)->uNewState & LVIS_STATEIMAGEMASK) >> 12) - 1;
          bOldState=((((NMLISTVIEW *)lParam)->uOldState & LVIS_STATEIMAGEMASK) >> 12) - 1;

          if (bNewState >=0 && bOldState >=0 && bNewState != bOldState)
            bListChanged=TRUE;
        }
        if (((NMLISTVIEW *)lParam)->uNewState & LVIS_SELECTED)
        {
          SendMessage(hWndHotkey, HKM_SETHOTKEY, ((NMLISTVIEW *)lParam)->lParam, 0);
          nSelItem=((NMLISTVIEW *)lParam)->iItem;

          if (nSelItem > 0)
          {
            EnableWindow(hWndPlayButton, TRUE);
            EnableWindow(hWndRepeat, TRUE);
            EnableWindow(hWndRepeatSpin, TRUE);
            EnableWindow(hWndRecordButton, FALSE);
            EnableWindow(hWndName, FALSE);
            EnableWindow(hWndSaveButton, FALSE);
            EnableWindow(hWndDeleteButton, TRUE);
            EnableWindow(hWndHotkey, TRUE);
            EnableWindow(hWndAssignButton, TRUE);
          }
          else
          {
            if (hRecordStack.last)
            {
              EnableWindow(hWndPlayButton, TRUE);
              EnableWindow(hWndRepeat, TRUE);
              EnableWindow(hWndRepeatSpin, TRUE);
              EnableWindow(hWndName, TRUE);
              if (GetWindowTextLength(hWndName))
                EnableWindow(hWndSaveButton, TRUE);
              else
                EnableWindow(hWndSaveButton, FALSE);
            }
            else
            {
              EnableWindow(hWndPlayButton, FALSE);
              EnableWindow(hWndRepeat, FALSE);
              EnableWindow(hWndRepeatSpin, FALSE);
              EnableWindow(hWndName, FALSE);
              EnableWindow(hWndSaveButton, FALSE);
            }
            EnableWindow(hWndRecordButton, TRUE);
            EnableWindow(hWndDeleteButton, FALSE);
            EnableWindow(hWndHotkey, FALSE);
            EnableWindow(hWndAssignButton, FALSE);
          }
        }
        if (((NMLISTVIEW *)lParam)->uOldState & LVIS_SELECTED)
        {
          SendMessage(hWndHotkey, HKM_SETHOTKEY, 0, 0);
          nSelItem=-1;

          EnableWindow(hWndPlayButton, FALSE);
          EnableWindow(hWndRepeat, FALSE);
          EnableWindow(hWndRepeatSpin, FALSE);
          EnableWindow(hWndRecordButton, FALSE);
          EnableWindow(hWndName, FALSE);
          EnableWindow(hWndSaveButton, FALSE);
          EnableWindow(hWndDeleteButton, FALSE);
          EnableWindow(hWndHotkey, FALSE);
          EnableWindow(hWndAssignButton, FALSE);
        }
      }
      else if ((int)((NMHDR *)lParam)->code == NM_DBLCLK) PostMessage(hDlg, WM_COMMAND, IDC_PLAY, 0);
    }
  }
  else if (uMsg == WM_COMMAND)
  {
    if (LOWORD(wParam) == IDC_PLAY)
    {
      wchar_t wszName[MAX_PATH];
      LVITEMW lvi;
      int nRepeat;

      if (nSelItem == 0)
      {
        EnableWindow(hWndPlayButton, FALSE);
        SetFocus(hMainWnd);
        nRepeat=GetDlgItemInt(hDlg, IDC_REPEAT, NULL, FALSE);
        MacroPlay(&hRecordStack, nRepeat);
        EnableWindow(hWndPlayButton, TRUE);
      }
      else
      {
        //Get macro name
        lvi.mask=LVIF_TEXT;
        lvi.pszText=wszName;
        lvi.cchTextMax=MAX_PATH;
        lvi.iItem=nSelItem;
        lvi.iSubItem=LVI_MACRO_FILE;
        SendMessage(hWndList, LVM_GETITEMW, 0, (LPARAM)&lvi);

        if (lstrcmpiW(wszName, wszLastMacro))
        {
          StackHotkeyFree(&hLastMacroStack);
          wszLastMacro[0]='\0';

          if (ReadMacroFileW(&hLastMacroStack, wszName))
            lstrcpynW(wszLastMacro, wszName, MAX_PATH);
        }
        EnableWindow(hWndPlayButton, FALSE);
        SetFocus(hMainWnd);
        nRepeat=GetDlgItemInt(hDlg, IDC_REPEAT, NULL, FALSE);
        MacroPlay(&hLastMacroStack, nRepeat);
        EnableWindow(hWndPlayButton, TRUE);
      }
    }
    else if (LOWORD(wParam) == IDC_RECORD)
    {
      if (!bRecord)
      {
        EnableWindow(hWndPlayButton, FALSE);
        EnableWindow(hWndRepeat, FALSE);
        EnableWindow(hWndRepeatSpin, FALSE);
        EnableWindow(hWndName, FALSE);
        EnableWindow(hWndSaveButton, FALSE);
        EnableWindow(hWndList, FALSE);

        StackHotkeyFree(&hRecordStack);
        SetWindowTextW(hWndRecordButton, GetLangStringW(wLangSystem, STRID_STOP));
        SetFocus(hMainWnd);
        MacroRecord();
        bRecord=TRUE;
      }
      else
      {
        bRecord=FALSE;
        MacroStop();
        SetWindowTextW(hWndRecordButton, GetLangStringW(wLangSystem, STRID_RECORD));

        EnableWindow(hWndList, TRUE);
        if (hRecordStack.last)
        {
          EnableWindow(hWndPlayButton, TRUE);
          EnableWindow(hWndRepeat, TRUE);
          EnableWindow(hWndRepeatSpin, TRUE);
          EnableWindow(hWndName, TRUE);
          if (GetWindowTextLengthW(hWndName))
            EnableWindow(hWndSaveButton, TRUE);
          else
            EnableWindow(hWndSaveButton, FALSE);
        }
      }
    }
    else if (LOWORD(wParam) == IDC_NAME)
    {
      if (GetWindowTextLengthW(hWndName))
        EnableWindow(hWndSaveButton, TRUE);
      else
        EnableWindow(hWndSaveButton, FALSE);
    }
    else if (LOWORD(wParam) == IDC_SAVE)
    {
      wchar_t wszName[MAX_PATH];
      LVITEMW lvi;
      int nIndex;

      if (GetWindowTextW(hWndName, wszName, MAX_PATH))
      {
        wsprintfW(wszBuffer, L"%s\\%s.macro", wszMacrosDir, wszName);

        if (!FileExistsW(wszBuffer))
        {
          if (SaveMacroFileW(&hRecordStack, wszName))
          {
            lstrcpynW(wszLastMacro, wszName, MAX_PATH);
            memcpy(&hLastMacroStack, &hRecordStack, sizeof(HSTACK));
            hRecordStack.first=0;
            hRecordStack.last=0;

            lvi.mask=LVIF_TEXT;
            lvi.pszText=wszName;
            lvi.iItem=SendMessage(hWndList, LVM_GETITEMCOUNT, 0, 0);
            lvi.iSubItem=LVI_MACRO_FILE;
            nIndex=SendMessage(hWndList, LVM_INSERTITEMW, 0, (LPARAM)&lvi);

            if (!nIndex) nIndex=MoveListViewItemW(hWndList, nIndex, nIndex + 1);

            SetFocus(hWndList);
            lvi.stateMask=LVIS_SELECTED|LVIS_FOCUSED;
            lvi.state=LVIS_SELECTED|LVIS_FOCUSED;
            SendMessage(hWndList, LVM_SETITEMSTATE, nIndex, (LPARAM)&lvi);
          }
          else MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_CANNOT_OPEN_FILE), L"Macros plugin", MB_OK|MB_ICONERROR);
        }
        else MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_ALREADY_EXIST), L"Macros plugin", MB_OK|MB_ICONEXCLAMATION);
      }
    }
    else if (LOWORD(wParam) == IDC_DELETE)
    {
      wchar_t wszName[MAX_PATH];
      LVITEMW lvi;

      if (nSelItem > 0)
      {
        if (MessageBoxW(hDlg, GetLangStringW(wLangSystem, STRID_CONFIRM_DELETE), L"Macros plugin", MB_YESNO|MB_ICONQUESTION) == IDYES)
        {
          //Get macro name
          lvi.mask=LVIF_TEXT;
          lvi.pszText=wszName;
          lvi.cchTextMax=MAX_PATH;
          lvi.iItem=nSelItem;
          lvi.iSubItem=LVI_MACRO_FILE;
          SendMessage(hWndList, LVM_GETITEMW, 0, (LPARAM)&lvi);

          if (DeleteMacroFileW(wszName))
          {
            if (!lstrcmpiW(wszName, wszLastMacro))
            {
              StackHotkeyFree(&hLastMacroStack);
              wszLastMacro[0]='\0';
            }
            SendMessage(hWndList, LVM_DELETEITEM, nSelItem, 0);
          }
        }
      }
    }
    else if (LOWORD(wParam) == IDC_ASSIGN)
    {
      LVITEMW lvi;
      WORD wHotkey;
      wchar_t wszFile[MAX_PATH];
      wchar_t wszHotkey[MAX_PATH];

      //Get macro name
      lvi.mask=LVIF_TEXT;
      lvi.pszText=wszFile;
      lvi.cchTextMax=MAX_PATH;
      lvi.iItem=nSelItem;
      lvi.iSubItem=LVI_MACRO_FILE;
      SendMessage(hWndList, LVM_GETITEMW, 0, (LPARAM)&lvi);

      //Get macro hotkey
      wHotkey=(WORD)SendMessage(hWndHotkey, HKM_GETHOTKEY, 0, 0);

      //Register hotkey
      if (RegisterHotkeyW(wszFile, wHotkey))
      {
        GetHotkeyStringW(wHotkey, wszHotkey);

        lvi.mask=LVIF_PARAM;
        lvi.iItem=nSelItem;
        lvi.iSubItem=LVI_MACRO_FILE;
        lvi.lParam=wHotkey;
        SendMessage(hWndList, LVM_SETITEMW, 0, (LPARAM)&lvi);

        lvi.mask=LVIF_TEXT;
        lvi.pszText=wszHotkey;
        lvi.iItem=nSelItem;
        lvi.iSubItem=LVI_MACRO_HOTKEY;
        SendMessage(hWndList, LVM_SETITEMW, 0, (LPARAM)&lvi);

        SendMessage(hWndHotkey, HKM_SETHOTKEY, wHotkey, 0);
        bListChanged=TRUE;
      }
      else SetFocus(hWndHotkey);
    }
    else if (LOWORD(wParam) == IDOK ||
             LOWORD(wParam) == IDCANCEL)
    {
      int nWidth;

      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_MACRO_FILE, 0);
      if (nWidth != nColumnWidth1)
      {
        nColumnWidth1=nWidth;
        bMainDialogRectSave=TRUE;
      }
      nWidth=SendMessage(hWndList, LVM_GETCOLUMNWIDTH, LVI_MACRO_HOTKEY, 0);
      if (nWidth != nColumnWidth2)
      {
        nColumnWidth2=nWidth;
        bMainDialogRectSave=TRUE;
      }

      if (bListChanged || LOWORD(wParam) == IDOK)
      {
        SaveMacroListW(hWndList);
        SaveOptionsW(OF_MAIN|OF_WINDOWRECT);
      }

      if (bRecord)
      {
        bRecord=FALSE;
        MacroStop();
      }
      DestroyIcon(hIconPlay);

      DestroyWindow(hWndMainDlg);
      hWndMainDlg=NULL;
      CloseHandle(hDialogThread);
      hDialogThread=NULL;
      ExitThread(0);
    }
  }
  else if (uMsg == WM_CLOSE)
  {
    PostMessage(hDlg, WM_COMMAND, IDCANCEL, 0);
    return TRUE;
  }

  if (DialogResizeMessages(&drs[0], &rcMainInitDialog, &rcMainCurrentDialog, hDlg, uMsg, wParam, lParam))
    bMainDialogRectSave=TRUE;

  return FALSE;
}

LRESULT CALLBACK GetMsgProc(int code, WPARAM wParam, LPARAM lParam)
{
  if (code >= 0)
  {
    if (wParam == PM_REMOVE)
    {
      MSG *msg=(MSG *)lParam;
      DWORD dwFlags=0;

      if (msg->message == WM_KEYDOWN || msg->message == WM_SYSKEYDOWN ||
          msg->message == WM_KEYUP || msg->message == WM_SYSKEYUP)
      {
        if ((msg->lParam >> 24) & 1)
          dwFlags|=KEYEVENTF_EXTENDEDKEY;
        if (msg->message == WM_KEYUP || msg->message == WM_SYSKEYUP)
          dwFlags|=KEYEVENTF_KEYUP;
        else if (msg->message == WM_KEYDOWN || msg->message == WM_SYSKEYDOWN)
        {
          if ((msg->lParam >> 30) & 1)
          {
            StackHotkeyAdd(&hRecordStack, msg->wParam, dwFlags|KEYEVENTF_KEYUP);
          }
        }
        StackHotkeyAdd(&hRecordStack, msg->wParam, dwFlags);
      }
    }
  }
  return CallNextHookEx(hHook, code, wParam, lParam);
}

BOOL RegisterHotkeyA(char *szMacroName, WORD wHotkey)
{
  PLUGINFUNCTIONA *pfElement=NULL;
  char szPrefix[MAX_PATH];
  char szFunction[MAX_PATH];

  lstrcpyA(szPrefix, "Macros::Main::");
  wsprintfA(szFunction, "%s%s", szPrefix, szMacroName);
  pfElement=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, wHotkey);

  if (!pfElement)
  {
    if (pfElement=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)szFunction, 0))
    {
      pfElement->wHotkey=wHotkey;

      if (!pfElement->wHotkey && !pfElement->bRunning && !pfElement->bOnStart)
        SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfElement);
    }
    else
    {
      if (wHotkey)
      {
        PLUGINFUNCTIONA pf={0};

        if (pfElement=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf))
        {
          lstrcpynA(pfElement->szFunction, szFunction, MAX_PATH);
          pfElement->nFunctionLen=lstrlenA(pfElement->szFunction);
          pfElement->wHotkey=wHotkey;
          pfElement->PluginProc=HotkeyProcA;
          pfElement->lpParameter=pfElement->szFunction + lstrlenA(szPrefix);
        }
      }
    }
    return TRUE;
  }
  else
  {
    wsprintfA(szBuffer, GetLangStringA(wLangSystem, STRID_HOTKEY_EXISTS), szFunction, pfElement->szFunction);
    MessageBoxA(hWndMainDlg, szBuffer, "Macros plugin", MB_OK|MB_ICONEXCLAMATION);
    return FALSE;
  }
}

BOOL RegisterHotkeyW(wchar_t *wszMacroName, WORD wHotkey)
{
  PLUGINFUNCTIONW *pfElement=NULL;
  wchar_t wszPrefix[MAX_PATH];
  wchar_t wszFunction[MAX_PATH];

  lstrcpyW(wszPrefix, L"Macros::Main::");
  wsprintfW(wszFunction, L"%s%s", wszPrefix, wszMacroName);
  pfElement=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)NULL, wHotkey);

  if (!pfElement)
  {
    if (pfElement=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)wszFunction, 0))
    {
      pfElement->wHotkey=wHotkey;

      if (!pfElement->wHotkey && !pfElement->bRunning && !pfElement->bOnStart)
        SendMessage(hMainWnd, AKD_DLLDELETE, 0, (LPARAM)pfElement);
    }
    else
    {
      if (wHotkey)
      {
        PLUGINFUNCTIONW pf={0};

        if (pfElement=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLADD, 0, (LPARAM)&pf))
        {
          lstrcpynW(pfElement->szFunction, wszFunction, MAX_PATH);
          pfElement->nFunctionLen=lstrlenW(pfElement->szFunction);
          pfElement->wHotkey=wHotkey;
          pfElement->PluginProc=HotkeyProcW;
          pfElement->lpParameter=pfElement->szFunction + lstrlenW(wszPrefix);
        }
      }
    }
    return TRUE;
  }
  else
  {
    wsprintfW(wszBuffer, GetLangStringW(wLangSystem, STRID_HOTKEY_EXISTS), wszFunction, pfElement->szFunction);
    MessageBoxW(hWndMainDlg, wszBuffer, L"Macros plugin", MB_OK|MB_ICONEXCLAMATION);
    return FALSE;
  }
}

BOOL CALLBACK HotkeyProcA(void *lpParameter)
{
  char *pName=(char *)lpParameter;

  ExecMacroA(pName, 1);
  return TRUE;
}

BOOL CALLBACK HotkeyProcW(void *lpParameter)
{
  wchar_t *wpName=(wchar_t *)lpParameter;

  ExecMacroW(wpName, 1);
  return TRUE;
}

DWORD WINAPI ExecThreadProcA(LPVOID lpParameter)
{
  EXECMACROA *em=(EXECMACROA *)lpParameter;
  HANDLE hInitMutex=em->hInitMutex;
  HANDLE hExecMutex;
  HANDLE hThread=hExecThread;
  char szMacro[MAX_PATH];
  int nRepeat;

  //Copy to thread memory
  lstrcpynA(szMacro, em->pMacro, MAX_PATH);
  nRepeat=em->nRepeat;

  //Protect from double execution
  hExecMutex=CreateEventA(NULL, FALSE, FALSE, szMutexExecName);

  //Thread is initialized now unlock main thread
  SetEvent(hInitMutex);

  if (hExecMutex)
  {
    if (GetLastError() == ERROR_ALREADY_EXISTS)
      WaitForSingleObject(hExecMutex, INFINITE);

    //Wait for release all virtual keys
    while (SendMessage(hMainWnd, AKD_WAITKEYBOARD, TRUE, 0))
      Sleep(100);

    if (!WaitIdle())
    {
      //Play macro
      if (lstrcmpiA(szMacro, szLastMacro))
      {
        StackHotkeyFree(&hLastMacroStack);
        szLastMacro[0]='\0';

        if (ReadMacroFileA(&hLastMacroStack, szMacro))
          lstrcpynA(szLastMacro, szMacro, MAX_PATH);
      }
      MacroPlay(&hLastMacroStack, nRepeat);
    }
    SetEvent(hExecMutex);
    CloseHandle(hExecMutex);
  }

  //Free thread handle
  if (hThread)
  {
    CloseHandle(hThread);
    hThread=NULL;
  }
  return 0;
}

DWORD WINAPI ExecThreadProcW(LPVOID lpParameter)
{
  EXECMACROW *em=(EXECMACROW *)lpParameter;
  HANDLE hInitMutex=em->hInitMutex;
  HANDLE hExecMutex;
  HANDLE hThread=hExecThread;
  wchar_t wszMacro[MAX_PATH];
  int nRepeat;

  //Copy to thread memory
  lstrcpynW(wszMacro, em->wpMacro, MAX_PATH);
  nRepeat=em->nRepeat;

  //Protect from double execution
  hExecMutex=CreateEventW(NULL, FALSE, FALSE, wszMutexExecName);

  //Thread is initialized now unlock main thread
  SetEvent(hInitMutex);

  if (hExecMutex)
  {
    if (GetLastError() == ERROR_ALREADY_EXISTS)
      WaitForSingleObject(hExecMutex, INFINITE);

    //Wait for release all virtual keys
    while (SendMessage(hMainWnd, AKD_WAITKEYBOARD, TRUE, 0))
      Sleep(100);

    if (!WaitIdle())
    {
      //Play macro
      if (lstrcmpiW(wszMacro, wszLastMacro))
      {
        StackHotkeyFree(&hLastMacroStack);
        wszLastMacro[0]='\0';

        if (ReadMacroFileW(&hLastMacroStack, wszMacro))
          lstrcpynW(wszLastMacro, wszMacro, MAX_PATH);
      }
      MacroPlay(&hLastMacroStack, nRepeat);
    }
    SetEvent(hExecMutex);
    CloseHandle(hExecMutex);
  }

  //Free thread handle
  if (hThread)
  {
    CloseHandle(hThread);
    hThread=NULL;
  }
  return 0;
}

KEYSTRUCT* StackHotkeyAdd(HSTACK *hStack, BYTE bVk, DWORD dwFlags)
{
  KEYSTRUCT *lpElement;

  if (!StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(KEYSTRUCT)))
  {
    lpElement->ka.bVk=bVk;
    lpElement->ka.dwFlags=dwFlags;
  }
  return lpElement;
}

void StackHotkeyPress(HSTACK *hStack, BOOL bToEnd)
{
  KEYSTRUCT *lpElement=(KEYSTRUCT *)hStack->first;
  HANDLE hScriptsExecMutex;
  BOOL bBusy=FALSE;

  while (lpElement)
  {
    if (!bPlay)
    {
      //Release key
      if (lpElement->ka.dwFlags & KEYEVENTF_KEYUP)
      {
        if (GetKeyState(lpElement->ka.bVk) < 0)
          keybd_event(lpElement->ka.bVk, 0, lpElement->ka.dwFlags, 0);
      }
    }
    else
    {
      //Press key
      keybd_event(lpElement->ka.bVk, 0, lpElement->ka.dwFlags, 0);
    }
    if (bBusy=WaitIdle())
      bPlay=FALSE;

    //Scripts plugin synchronization
    while (hScriptsExecMutex=OpenEventA(EVENT_ALL_ACCESS, FALSE, "AkelPad::Scripts::MutexExec"))
    {
      WaitForSingleObject(hScriptsExecMutex, INFINITE);

      if (bBusy=WaitIdle())
        bPlay=FALSE;

      SetEvent(hScriptsExecMutex);
      CloseHandle(hScriptsExecMutex);
    }

    //Check is end of the file reached
    if (bToEnd && bPlay)
    {
      EDITINFO ei;
      CHARRANGE chrg;
      TEXTRANGEA txtrng;
      char szEnd[4];

      if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
      {
        SendMessage(ei.hWndEdit, EM_EXGETSEL, 0, (LPARAM)&chrg);
        txtrng.chrg.cpMin=chrg.cpMax;
        txtrng.chrg.cpMax=chrg.cpMax + 1;
        txtrng.lpstrText=szEnd;
        SendMessageA(ei.hWndEdit, EM_GETTEXTRANGE, 0, (LPARAM)&txtrng);
        if (*szEnd == '\0') bPlay=FALSE;
      }
    }
    lpElement=lpElement->next;
  }

  if (bBusy)
    MessageBoxW(hMainWnd, GetLangStringW(wLangSystem, STRID_THREAD_BUSY), L"Macros plugin", MB_OK|MB_ICONEXCLAMATION);
}

BOOL StackHotkeyRead(HSTACK *hStack, HANDLE hFile)
{
  KEYSTRUCT *lpElement;
  KEYACT ka;
  DWORD dwBytesRead;

  while (1)
  {
    if (!ReadFile(hFile, &ka, sizeof(KEYACT), &dwBytesRead, NULL))
      return FALSE;
    if (!dwBytesRead)
      break;
    if (!StackInsertIndex((stack **)&hStack->first, (stack **)&hStack->last, (stack **)&lpElement, -1, sizeof(KEYSTRUCT)))
      memcpy(&lpElement->ka, &ka, sizeof(KEYACT));
    else
      return FALSE;
  }
  return TRUE;
}

BOOL StackHotkeySave(HSTACK *hStack, HANDLE hFile)
{
  KEYSTRUCT *lpElement=(KEYSTRUCT *)hStack->first;
  DWORD dwBytesWritten;

  while (lpElement)
  {
    if (!WriteFile(hFile, &lpElement->ka, sizeof(KEYACT), &dwBytesWritten, NULL))
      return FALSE;

    lpElement=lpElement->next;
  }
  return TRUE;
}

void StackHotkeyFree(HSTACK *hStack)
{
  StackClear((stack **)&hStack->first, (stack **)&hStack->last);
}

BOOL ReadMacroFileA(HSTACK *hKeyStack, char *pMacro)
{
  HANDLE hFile;
  char szMacro[MAX_PATH];

  wsprintfA(szMacro, "%s\\%s.macro", szMacrosDir, pMacro);

  if ((hFile=CreateFileA(szMacro, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) != INVALID_HANDLE_VALUE)
  {
    StackHotkeyRead(hKeyStack, hFile);
    CloseHandle(hFile);
    return TRUE;
  }
  return FALSE;
}

BOOL ReadMacroFileW(HSTACK *hKeyStack, wchar_t *wpMacro)
{
  HANDLE hFile;
  wchar_t wszMacro[MAX_PATH];

  wsprintfW(wszMacro, L"%s\\%s.macro", wszMacrosDir, wpMacro);

  if ((hFile=CreateFileW(wszMacro, GENERIC_READ, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) != INVALID_HANDLE_VALUE)
  {
    StackHotkeyRead(hKeyStack, hFile);
    CloseHandle(hFile);
    return TRUE;
  }
  return FALSE;
}

BOOL SaveMacroFileA(HSTACK *hRecordStack, char *pMacro)
{
  HANDLE hFile;
  char szMacro[MAX_PATH];

  wsprintfA(szMacro, "%s\\%s.macro", szMacrosDir, pMacro);
  CreateDirectoryA(szMacrosDir, NULL);

  if ((hFile=CreateFileA(szMacro, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL)) != INVALID_HANDLE_VALUE)
  {
    StackHotkeySave(hRecordStack, hFile);
    CloseHandle(hFile);
    return TRUE;
  }
  return FALSE;
}

BOOL SaveMacroFileW(HSTACK *hRecordStack, wchar_t *wpMacro)
{
  HANDLE hFile;
  wchar_t wszMacro[MAX_PATH];

  wsprintfW(wszMacro, L"%s\\%s.macro", wszMacrosDir, wpMacro);
  CreateDirectoryW(wszMacrosDir, NULL);

  if ((hFile=CreateFileW(wszMacro, GENERIC_WRITE, FILE_SHARE_READ|FILE_SHARE_WRITE, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL)) != INVALID_HANDLE_VALUE)
  {
    StackHotkeySave(hRecordStack, hFile);
    CloseHandle(hFile);
    return TRUE;
  }
  return FALSE;
}

BOOL DeleteMacroFileA(char *pMacro)
{
  char szMacro[MAX_PATH];

  wsprintfA(szMacro, "%s\\%s.macro", szMacrosDir, pMacro);
  return DeleteFileA(szMacro);
}

BOOL DeleteMacroFileW(wchar_t *wpMacro)
{
  wchar_t wszMacro[MAX_PATH];

  wsprintfW(wszMacro, L"%s\\%s.macro", wszMacrosDir, wpMacro);
  return DeleteFileW(wszMacro);
}

void FillMacroListA(HWND hWnd)
{
  WIN32_FIND_DATAA wfd;
  HANDLE hFind;
  char szFindFiles[MAX_PATH];
  char szBaseName[MAX_PATH];
  char szHotkey[MAX_PATH];
  LVITEMA lvi;
  DWORD dwStyle;
  int nIndex;
  int i=0;

  //Macro files
  wsprintfA(szFindFiles, "%s\\*.macro", szMacrosDir);

  if ((hFind=FindFirstFileA(szFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
  {
    do
    {
      GetBaseNameA(wfd.cFileName, szBaseName, MAX_PATH);

      //Find hotkey
      {
        PLUGINFUNCTIONA *pfElement=NULL;

        lvi.mask=LVIF_TEXT;
        lvi.pszText=szBaseName;
        lvi.iItem=i++;
        lvi.iSubItem=LVI_MACRO_FILE;
        nIndex=SendMessage(hWnd, LVM_INSERTITEMA, 0, (LPARAM)&lvi);

        wsprintfA(szBuffer, "Macros::Main::%s", szBaseName);
        pfElement=(PLUGINFUNCTIONA *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)szBuffer, 0);

        if (pfElement)
        {
          lvi.mask=LVIF_PARAM;
          lvi.iItem=nIndex;
          lvi.iSubItem=LVI_MACRO_FILE;
          lvi.lParam=pfElement->wHotkey;
          SendMessage(hWnd, LVM_SETITEMA, 0, (LPARAM)&lvi);

          GetHotkeyStringA(pfElement->wHotkey, szHotkey);
          lvi.mask=LVIF_TEXT;
          lvi.pszText=szHotkey;
          lvi.iItem=nIndex;
          lvi.iSubItem=LVI_MACRO_HOTKEY;
          SendMessage(hWnd, LVM_SETITEMA, 0, (LPARAM)&lvi);
        }
      }
    }
    while (FindNextFileA(hFind, &wfd));

    FindClose(hFind);
  }

  dwStyle=GetWindowLongA(hWnd, GWL_STYLE);
  SetWindowLongA(hWnd, GWL_STYLE, dwStyle & ~LVS_SORTASCENDING);
  lvi.mask=LVIF_TEXT;
  lvi.pszText=GetLangStringA(wLangSystem, STRID_RECORD_LISTITEM);
  lvi.iItem=0;
  lvi.iSubItem=LVI_MACRO_FILE;
  nIndex=SendMessage(hWnd, LVM_INSERTITEMA, 0, (LPARAM)&lvi);
  SetWindowLongA(hWnd, GWL_STYLE, dwStyle);

  lvi.stateMask=LVIS_SELECTED|LVIS_FOCUSED;
  lvi.state=LVIS_SELECTED|LVIS_FOCUSED;
  SendMessage(hWnd, LVM_SETITEMSTATE, 0, (LPARAM)&lvi);
}

void FillMacroListW(HWND hWnd)
{
  WIN32_FIND_DATAW wfd;
  HANDLE hFind;
  wchar_t wszFindFiles[MAX_PATH];
  wchar_t wszBaseName[MAX_PATH];
  wchar_t wszHotkey[MAX_PATH];
  LVITEMW lvi;
  DWORD dwStyle;
  int nIndex;
  int i=0;

  //Macro files
  wsprintfW(wszFindFiles, L"%s\\*.macro", wszMacrosDir);

  if ((hFind=FindFirstFileW(wszFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
  {
    do
    {
      GetBaseNameW(wfd.cFileName, wszBaseName, MAX_PATH);

      //Find hotkey
      {
        PLUGINFUNCTIONW *pfElement=NULL;

        lvi.mask=LVIF_TEXT;
        lvi.pszText=wszBaseName;
        lvi.iItem=i++;
        lvi.iSubItem=LVI_MACRO_FILE;
        nIndex=SendMessage(hWnd, LVM_INSERTITEMW, 0, (LPARAM)&lvi);

        wsprintfW(wszBuffer, L"Macros::Main::%s", wszBaseName);
        pfElement=(PLUGINFUNCTIONW *)SendMessage(hMainWnd, AKD_DLLFIND, (WPARAM)wszBuffer, 0);

        if (pfElement)
        {
          lvi.mask=LVIF_PARAM;
          lvi.iItem=nIndex;
          lvi.iSubItem=LVI_MACRO_FILE;
          lvi.lParam=pfElement->wHotkey;
          SendMessage(hWnd, LVM_SETITEMW, 0, (LPARAM)&lvi);

          GetHotkeyStringW(pfElement->wHotkey, wszHotkey);
          lvi.mask=LVIF_TEXT;
          lvi.pszText=wszHotkey;
          lvi.iItem=nIndex;
          lvi.iSubItem=LVI_MACRO_HOTKEY;
          SendMessage(hWnd, LVM_SETITEMW, 0, (LPARAM)&lvi);
        }
      }
    }
    while (FindNextFileW(hFind, &wfd));

    FindClose(hFind);
  }

  dwStyle=GetWindowLongW(hWnd, GWL_STYLE);
  SetWindowLongW(hWnd, GWL_STYLE, dwStyle & ~LVS_SORTASCENDING);
  lvi.mask=LVIF_TEXT;
  lvi.pszText=GetLangStringW(wLangSystem, STRID_RECORD_LISTITEM);
  lvi.iItem=0;
  lvi.iSubItem=LVI_MACRO_FILE;
  nIndex=SendMessage(hWnd, LVM_INSERTITEMW, 0, (LPARAM)&lvi);
  SetWindowLongW(hWnd, GWL_STYLE, dwStyle);

  lvi.stateMask=LVIS_SELECTED|LVIS_FOCUSED;
  lvi.state=LVIS_SELECTED|LVIS_FOCUSED;
  SendMessage(hWnd, LVM_SETITEMSTATE, 0, (LPARAM)&lvi);
}

void SaveMacroListA(HWND hWnd)
{
  PLUGINOPTIONA po;
  HANDLE hOptions;
  char szFile[MAX_PATH];
  LVITEMA lvi;
  int i=1;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE|POB_CLEAR, (LPARAM)szPluginName))
  {
    while (1)
    {
      //Get macro name
      lvi.mask=LVIF_TEXT;
      lvi.pszText=szFile;
      lvi.cchTextMax=MAX_PATH;
      lvi.iItem=i;
      lvi.iSubItem=LVI_MACRO_FILE;
      if (!SendMessage(hWnd, LVM_GETITEMA, 0, (LPARAM)&lvi)) break;

      //Get macro hotkey
      lvi.mask=LVIF_PARAM;
      lvi.iItem=i;
      lvi.iSubItem=LVI_MACRO_FILE;
      if (!SendMessage(hWnd, LVM_GETITEMA, 0, (LPARAM)&lvi)) break;

      //Save hotkey
      if (lvi.lParam)
      {
        po.pOptionName=szFile;
        po.dwType=PO_DWORD;
        po.lpData=(LPBYTE)&lvi.lParam;
        po.dwData=sizeof(DWORD);
        SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
      }

      ++i;
    }
    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveMacroListW(HWND hWnd)
{
  PLUGINOPTIONW po;
  HANDLE hOptions;
  wchar_t wszFile[MAX_PATH];
  LVITEMW lvi;
  int i=1;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE|POB_CLEAR, (LPARAM)wszPluginName))
  {
    while (1)
    {
      //Get macro name
      lvi.mask=LVIF_TEXT;
      lvi.pszText=wszFile;
      lvi.cchTextMax=MAX_PATH;
      lvi.iItem=i;
      lvi.iSubItem=LVI_MACRO_FILE;
      if (!SendMessage(hWnd, LVM_GETITEMW, 0, (LPARAM)&lvi)) break;

      //Get macro hotkey
      lvi.mask=LVIF_PARAM;
      lvi.iItem=i;
      lvi.iSubItem=LVI_MACRO_FILE;
      if (!SendMessage(hWnd, LVM_GETITEMW, 0, (LPARAM)&lvi)) break;

      //Save hotkey
      if (lvi.lParam)
      {
        po.pOptionName=wszFile;
        po.dwType=PO_DWORD;
        po.lpData=(LPBYTE)&lvi.lParam;
        po.dwData=sizeof(DWORD);
        SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
      }

      ++i;
    }
    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void MacroRecord()
{
  hHook=SetWindowsHookEx(WH_GETMESSAGE, GetMsgProc, hInstanceDLL, 0);
}

void MacroStop()
{
  if (hHook)
  {
    if (UnhookWindowsHookEx(hHook))
      hHook=NULL;
  }
}

void MacroPlay(HSTACK *hStack, int nRepeat)
{
  EDITINFO ei;
  DWORD dwWindowProcessId;
  DWORD dwWindowThreadId;

  if (SendMessage(hMainWnd, AKD_GETEDITINFO, (WPARAM)NULL, (LPARAM)&ei))
  {
    if (!ei.bReadOnly)
    {
      //Play
      bPlay=TRUE;

      do
      {
        dwWindowThreadId=GetWindowThreadProcessId(GetForegroundWindow(), &dwWindowProcessId);
        if (dwWindowProcessId != dwMainProcessId) break;
        if (hDialogThread && dwWindowThreadId == dwDialogThreadId) break;
        StackHotkeyPress(hStack, !nRepeat);
        if (!bPlay) break;
      }
      while (nRepeat == 0 || --nRepeat > 0);
    }
    else MessageBeep(MB_OK);
  }
}

BOOL WaitIdle()
{
  MSG msg;
  DWORD dwTypes;
  DWORD dwStartWait;

  //Process plugin thread messages
  if (hDialogThread)
  {
    while (PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
    {
      TranslateMessage(&msg);
      DispatchMessage(&msg);
    }
  }

  //Wait for process main thread messages
  dwStartWait=GetTickCount();

  while (1)
  {
    if (GetTickCount() - dwStartWait < MACRO_TIMEOUT)
    {
      Sleep(0);
      dwTypes=SendMessage(hMainWnd, AKD_GETQUEUE, //QS_KEY|
                                                  QS_MOUSEMOVE|
                                                  QS_MOUSEBUTTON|
                                                  QS_POSTMESSAGE|
                                                  QS_TIMER|
                                                  QS_PAINT|
                                                  QS_SENDMESSAGE|
                                                  QS_HOTKEY, 0);
      if (dwTypes == 0) break;
    }
    else return TRUE;
  }
  return FALSE;
}

void ExecMacroA(char *pMacro, int nRepeat)
{
  EXECMACROA em;
  HANDLE hInitMutex;

  if (!hDialogThread)
  {
    if (hInitMutex=CreateEventA(NULL, FALSE, FALSE, szMutexInitName))
    {
      em.pMacro=pMacro;
      em.nRepeat=nRepeat;
      em.hInitMutex=hInitMutex;

      if (hExecThread=CreateThread(NULL, 0, ExecThreadProcA, &em, 0, &dwExecThreadId))
        WaitForSingleObject(hInitMutex, INFINITE);

      CloseHandle(hInitMutex);
    }
  }
}

void ExecMacroW(wchar_t *wpMacro, int nRepeat)
{
  EXECMACROW em;
  HANDLE hInitMutex;

  if (!hDialogThread)
  {
    if (hInitMutex=CreateEventW(NULL, FALSE, FALSE, wszMutexInitName))
    {
      em.wpMacro=wpMacro;
      em.nRepeat=nRepeat;
      em.hInitMutex=hInitMutex;

      if (hExecThread=CreateThread(NULL, 0, ExecThreadProcW, &em, 0, &dwExecThreadId))
        WaitForSingleObject(hInitMutex, INFINITE);

      CloseHandle(hInitMutex);
    }
  }
}

LRESULT CALLBACK NewMainProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == AKDN_SEARCH_ENDED ||
      uMsg == AKDN_FRAME_NOWINDOWS)
  {
    bPlay=FALSE;
  }
  else if (uMsg == AKDN_MAIN_ONFINISH)
  {
    bPlay=FALSE;
    NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
    if (hWndMainDlg) SendMessage(hWndMainDlg, WM_COMMAND, IDCANCEL, 0);
    return FALSE;
  }

  //Call next procedure
  return NewMainProcData->NextProc(hWnd, uMsg, wParam, lParam);
}

void GetHotkeyStringA(WORD wHotkey, char *szString)
{
  char szKeyText[100];
  LONG lScan;

  szString[0]='\0';

  if (wHotkey)
  {
    if (HIBYTE(wHotkey) & HOTKEYF_CONTROL)
      lstrcatA(szString, "Ctrl + ");
    if (HIBYTE(wHotkey) & HOTKEYF_SHIFT)
      lstrcatA(szString, "Shift + ");
    if (HIBYTE(wHotkey) & HOTKEYF_ALT)
      lstrcatA(szString, "Alt + ");
    lScan=MapVirtualKey(LOBYTE(wHotkey), 0) << 16;
    if (HIBYTE(wHotkey) & HOTKEYF_EXT) lScan|=0x01000000L;
    GetKeyNameTextA(lScan, szKeyText, 100);
    lstrcatA(szString, szKeyText);
  }
}

void GetHotkeyStringW(WORD wHotkey, wchar_t *wszString)
{
  wchar_t wszKeyText[100];
  LONG lScan;

  wszString[0]='\0';

  if (wHotkey)
  {
    if (HIBYTE(wHotkey) & HOTKEYF_CONTROL)
      lstrcatW(wszString, L"Ctrl + ");
    if (HIBYTE(wHotkey) & HOTKEYF_SHIFT)
      lstrcatW(wszString, L"Shift + ");
    if (HIBYTE(wHotkey) & HOTKEYF_ALT)
      lstrcatW(wszString, L"Alt + ");
    lScan=MapVirtualKeyW(LOBYTE(wHotkey), 0) << 16;
    if (HIBYTE(wHotkey) & HOTKEYF_EXT) lScan|=0x01000000L;
    GetKeyNameTextW(lScan, wszKeyText, 100);
    lstrcatW(wszString, wszKeyText);
  }
}

int MoveListViewItemA(HWND hWnd, int nOldIndex, int nNewIndex)
{
  char szText[MAX_PATH];
  LVITEMA lvi;
  DWORD dwStyle;
  int nIndex;

  //Get item text
  lvi.mask=LVIF_TEXT;
  lvi.pszText=szText;
  lvi.cchTextMax=MAX_PATH;
  lvi.iItem=nOldIndex;
  lvi.iSubItem=0;
  SendMessage(hWnd, LVM_GETITEMA, 0, (LPARAM)&lvi);

  //Delete item
  SendMessage(hWnd, LVM_DELETEITEM, nOldIndex, 0);

  //Insert item
  dwStyle=GetWindowLongA(hWnd, GWL_STYLE);
  SetWindowLongA(hWnd, GWL_STYLE, dwStyle & ~LVS_SORTASCENDING);

  lvi.mask=LVIF_TEXT;
  lvi.pszText=szText;
  lvi.iItem=nNewIndex;
  lvi.iSubItem=0;
  nIndex=SendMessage(hWnd, LVM_INSERTITEMA, 0, (LPARAM)&lvi);

  SetWindowLongA(hWnd, GWL_STYLE, dwStyle);

  return nIndex;
}

int MoveListViewItemW(HWND hWnd, int nOldIndex, int nNewIndex)
{
  wchar_t wszText[MAX_PATH];
  LVITEMW lvi;
  DWORD dwStyle;
  int nIndex;

  //Get item text
  lvi.mask=LVIF_TEXT;
  lvi.pszText=wszText;
  lvi.cchTextMax=MAX_PATH;
  lvi.iItem=nOldIndex;
  lvi.iSubItem=0;
  SendMessage(hWnd, LVM_GETITEMW, 0, (LPARAM)&lvi);

  //Delete item
  SendMessage(hWnd, LVM_DELETEITEM, nOldIndex, 0);

  //Insert item
  dwStyle=GetWindowLongW(hWnd, GWL_STYLE);
  SetWindowLongW(hWnd, GWL_STYLE, dwStyle & ~LVS_SORTASCENDING);

  lvi.mask=LVIF_TEXT;
  lvi.pszText=wszText;
  lvi.iItem=nNewIndex;
  lvi.iSubItem=0;
  nIndex=SendMessage(hWnd, LVM_INSERTITEMW, 0, (LPARAM)&lvi);

  SetWindowLongW(hWnd, GWL_STYLE, dwStyle);

  return nIndex;
}

int GetBaseNameA(char *pFile, char *szBaseName, int nBaseNameMaxLen)
{
  int nFileLen=lstrlenA(pFile);
  int nEndOffset=-1;
  int i;

  for (i=nFileLen - 1; i >= 0; --i)
  {
    if (pFile[i] == '\\')
      break;

    if (nEndOffset == -1)
    {
      if (pFile[i] == '.')
        nEndOffset=i;
    }
  }
  ++i;
  if (nEndOffset == -1) nEndOffset=nFileLen;
  nBaseNameMaxLen=min(nEndOffset - i + 1, nBaseNameMaxLen);
  lstrcpynA(szBaseName, pFile + i, nBaseNameMaxLen);

  return nBaseNameMaxLen;
}

int GetBaseNameW(wchar_t *wpFile, wchar_t *wszBaseName, int nBaseNameMaxLen)
{
  int nFileLen=lstrlenW(wpFile);
  int nEndOffset=-1;
  int i;

  for (i=nFileLen - 1; i >= 0; --i)
  {
    if (wpFile[i] == '\\')
      break;

    if (nEndOffset == -1)
    {
      if (wpFile[i] == '.')
        nEndOffset=i;
    }
  }
  ++i;
  if (nEndOffset == -1) nEndOffset=nFileLen;
  nBaseNameMaxLen=min(nEndOffset - i + 1, nBaseNameMaxLen);
  lstrcpynW(wszBaseName, wpFile + i, nBaseNameMaxLen);

  return nBaseNameMaxLen;
}

BOOL FileExistsA(char *pFile)
{
  WIN32_FIND_DATAA wfd;
  HANDLE hFind;

  if ((hFind=FindFirstFileA(pFile, &wfd)) == INVALID_HANDLE_VALUE)
    return FALSE;

  FindClose(hFind);
  return TRUE;
}

BOOL FileExistsW(wchar_t *wpFile)
{
  WIN32_FIND_DATAW wfd;
  HANDLE hFind;

  if ((hFind=FindFirstFileW(wpFile, &wfd)) == INVALID_HANDLE_VALUE)
    return FALSE;

  FindClose(hFind);
  return TRUE;
}

BOOL GetWindowPos(HWND hWnd, HWND hWndOwner, RECT *rc)
{
  if (GetWindowRect(hWnd, rc))
  {
    rc->right-=rc->left;
    rc->bottom-=rc->top;

    if (hWndOwner)
    {
      if (!ScreenToClient(hWndOwner, (POINT *)&rc->left))
        return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

BOOL DialogResizeMessages(DIALOGRESIZE *drs, RECT *rcInit, RECT *rcCurrent, HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  if (uMsg == WM_INITDIALOG)
  {
    RECT rcTemplate;

    GetWindowPos(hDlg, NULL, rcInit);
    rcTemplate=*rcCurrent;
    *rcCurrent=*rcInit;

    if (rcTemplate.right && rcTemplate.bottom)
    {
      rcTemplate.left=rcInit->left + (rcInit->right - rcTemplate.right) / 2;
      rcTemplate.top=rcInit->top + (rcInit->bottom - rcTemplate.bottom) / 2;
      SetWindowPos(hDlg, 0, rcTemplate.left, rcTemplate.top, rcTemplate.right, rcTemplate.bottom, SWP_NOZORDER);
    }
  }
  else if (uMsg == WM_GETMINMAXINFO)
  {
    MINMAXINFO *mmi=(MINMAXINFO *)lParam;

    mmi->ptMinTrackSize.x=rcInit->right;
    mmi->ptMinTrackSize.y=rcInit->bottom;
  }
  else if (uMsg == WM_SIZE)
  {
    if (lParam)
    {
      RECT rcTemplate;
      RECT rcControl;
      DWORD dwFlags;
      POINT pt;
      int i;

      GetWindowPos(hDlg, NULL, &rcTemplate);
      pt.x=rcTemplate.right - rcCurrent->right;
      pt.y=rcTemplate.bottom - rcCurrent->bottom;
      *rcCurrent=rcTemplate;

      for (i=0; drs[i].lpWnd; ++i)
      {
        if (*drs[i].lpWnd)
        {
          dwFlags=0;
          if (drs[i].dwType & DRS_SIZE)
            dwFlags|=SWP_NOMOVE;
          else if (drs[i].dwType & DRS_MOVE)
            dwFlags|=SWP_NOSIZE;
          GetWindowPos(*drs[i].lpWnd, hDlg, &rcControl);
          SetWindowPos(*drs[i].lpWnd, 0, rcControl.left + ((drs[i].dwType & DRS_X)?pt.x:0), rcControl.top + ((drs[i].dwType & DRS_Y)?pt.y:0), rcControl.right + ((drs[i].dwType & DRS_X)?pt.x:0), rcControl.bottom + ((drs[i].dwType & DRS_Y)?pt.y:0), dwFlags|SWP_NOZORDER);
        }
      }
      InvalidateRect(hDlg, NULL, TRUE);
      return TRUE;
    }
  }
  else if (uMsg == WM_PAINT)
  {
    PAINTSTRUCT ps;
    RECT rcGrip;
    HDC hDC;

    if (hDC=BeginPaint(hDlg, &ps))
    {
      GetClientRect(hDlg, &rcGrip);
      rcGrip.left=rcGrip.right - GetSystemMetrics(SM_CXVSCROLL);
      rcGrip.top=rcGrip.bottom - GetSystemMetrics(SM_CYVSCROLL);
      DrawFrameControl(hDC, &rcGrip, DFC_SCROLL, DFCS_SCROLLSIZEGRIP);
      EndPaint(hDlg, &ps);
    }
  }
  return FALSE;
}


//// Options

void ReadOptionsA()
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)szPluginName))
  {
    po.pOptionName="/WindowRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcMainCurrentDialog;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="/ColumnWidth1";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth1;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName="/ColumnWidth2";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth2;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    //Macro hotkeys
    {
      WIN32_FIND_DATAA wfd;
      HANDLE hFind;
      char szFindFiles[MAX_PATH];
      char szBaseName[MAX_PATH];

      wsprintfA(szFindFiles, "%s\\*.macro", szMacrosDir);

      if ((hFind=FindFirstFileA(szFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
      {
        do
        {
          GetBaseNameA(wfd.cFileName, szBaseName, MAX_PATH);

          //Read and register hotkey
          {
            PLUGINOPTIONA po;
            DWORD dwHotkey=0;

            if (hOptions)
            {
              po.pOptionName=szBaseName;
              po.dwType=PO_DWORD;
              po.lpData=(LPBYTE)&dwHotkey;
              po.dwData=sizeof(DWORD);
              SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
            }
            if (dwHotkey) RegisterHotkeyA(szBaseName, LOWORD(dwHotkey));
          }
        }
        while (FindNextFileA(hFind, &wfd));

        FindClose(hFind);
      }
    }
    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void ReadOptionsW()
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_READ, (LPARAM)wszPluginName))
  {
    po.pOptionName=L"/WindowRect";
    po.dwType=PO_BINARY;
    po.lpData=(LPBYTE)&rcMainCurrentDialog;
    po.dwData=sizeof(RECT);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"/ColumnWidth1";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth1;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    po.pOptionName=L"/ColumnWidth2";
    po.dwType=PO_DWORD;
    po.lpData=(LPBYTE)&nColumnWidth2;
    po.dwData=sizeof(DWORD);
    SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

    //Macro hotkeys
    {
      WIN32_FIND_DATAW wfd;
      HANDLE hFind;
      wchar_t wszFindFiles[MAX_PATH];
      wchar_t wszBaseName[MAX_PATH];

      wsprintfW(wszFindFiles, L"%s\\*.macro", wszMacrosDir);

      if ((hFind=FindFirstFileW(wszFindFiles, &wfd)) != INVALID_HANDLE_VALUE)
      {
        do
        {
          GetBaseNameW(wfd.cFileName, wszBaseName, MAX_PATH);

          //Read and register hotkey
          {
            PLUGINOPTIONW po;
            DWORD dwHotkey=0;

            if (hOptions)
            {
              po.pOptionName=wszBaseName;
              po.dwType=PO_DWORD;
              po.lpData=(LPBYTE)&dwHotkey;
              po.dwData=sizeof(DWORD);
              SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
            }
            if (dwHotkey) RegisterHotkeyW(wszBaseName, LOWORD(dwHotkey));
          }
        }
        while (FindNextFileW(hFind, &wfd));

        FindClose(hFind);
      }
    }
    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsA(DWORD dwFlags)
{
  PLUGINOPTIONA po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)szPluginName))
  {
    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName="/WindowRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&rcMainCurrentDialog;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="/ColumnWidth1";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth1;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName="/ColumnWidth2";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth2;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

void SaveOptionsW(DWORD dwFlags)
{
  PLUGINOPTIONW po;
  HANDLE hOptions;

  if (hOptions=(HANDLE)SendMessage(hMainWnd, AKD_BEGINOPTIONS, POB_SAVE, (LPARAM)wszPluginName))
  {
    if (dwFlags & OF_WINDOWRECT)
    {
      po.pOptionName=L"/WindowRect";
      po.dwType=PO_BINARY;
      po.lpData=(LPBYTE)&rcMainCurrentDialog;
      po.dwData=sizeof(RECT);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"/ColumnWidth1";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth1;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);

      po.pOptionName=L"/ColumnWidth2";
      po.dwType=PO_DWORD;
      po.lpData=(LPBYTE)&nColumnWidth2;
      po.dwData=sizeof(DWORD);
      SendMessage(hMainWnd, AKD_OPTION, (WPARAM)hOptions, (LPARAM)&po);
    }

    SendMessage(hMainWnd, AKD_ENDOPTIONS, (WPARAM)hOptions, 0);
  }
}

char* GetLangStringA(LANGID wLangID, int nStringID)
{
  static char szStringBuf[MAX_PATH];

  WideCharToMultiByte(CP_ACP, 0, GetLangStringW(wLangID, nStringID), -1, szStringBuf, MAX_PATH, NULL, NULL);
  return szStringBuf;
}

wchar_t* GetLangStringW(LANGID wLangID, int nStringID)
{
  if (wLangID == LANG_RUSSIAN)
  {
    if (nStringID == STRID_HOTKEY_EXISTS)
      return L"%s\n\n\x0413\x043E\x0440\x044F\x0447\x0430\x044F\x0020\x043A\x043B\x0430\x0432\x0438\x0448\x0430\x0020\x0443\x0436\x0435\x0020\x043D\x0430\x0437\x043D\x0430\x0447\x0435\x043D\x0430\x0020\x0434\x043B\x044F\x0020\x0022\x0025\x0073\x0022\x002E";
    if (nStringID == STRID_CANNOT_OPEN_FILE)
      return L"\x041D\x0435\x0432\x043E\x0437\x043C\x043E\x0436\x043D\x043E\x0020\x043E\x0442\x043A\x0440\x044B\x0442\x044C\x0020\x0444\x0430\x0439\x043B\x002E";
    if (nStringID == STRID_ALREADY_EXIST)
      return L"\x041C\x0430\x043A\x0440\x043E\x0441\x0020\x0443\x0436\x0435\x0020\x0441\x0443\x0449\x0435\x0441\x0442\x0432\x0443\x0435\x0442\x002E";
    if (nStringID == STRID_CONFIRM_DELETE)
      return L"\x0412\x044B\x0020\x0443\x0432\x0435\x0440\x0435\x043D\x044B\x002C\x0020\x0447\x0442\x043E\x0020\x0445\x043E\x0442\x0438\x0442\x0435\x0020\x0443\x0434\x0430\x043B\x0438\x0442\x044C\x0020\x043C\x0430\x043A\x0440\x043E\x0441\x003F";
    if (nStringID == STRID_THREAD_BUSY)
      return L"\x0413\x043B\x0430\x0432\x043D\x044B\x0439\x0020\x043F\x043E\x0442\x043E\x043A\x0020\x0437\x0430\x043D\x044F\x0442\x002E";
    if (nStringID == STRID_MACRO)
      return L"\x041C\x0430\x043A\x0440\x043E\x0441";
    if (nStringID == STRID_HOTKEY)
      return L"\x0413\x043E\x0440\x044F\x0447\x0430\x044F\x0020\x043A\x043B\x0430\x0432\x0438\x0448\x0430";
    if (nStringID == STRID_RECORD_LISTITEM)
      return L"\x003C\x0417\x0430\x043F\x0438\x0441\x044C\x003E";
    if (nStringID == STRID_RECORD)
      return L"\x0417\x0430\x043F\x0438\x0441\x044C";
    if (nStringID == STRID_STOP)
      return L"\x0421\x0442\x043E\x043F";
    if (nStringID == STRID_SAVE)
      return L"\x0421\x043E\x0445\x0440\x0430\x043D\x0438\x0442\x044C";
    if (nStringID == STRID_DELETE)
      return L"\x0423\x0434\x0430\x043B\x0438\x0442\x044C";
    if (nStringID == STRID_ASSIGN)
      return L"\x041D\x0430\x0437\x043D\x0430\x0447\x0438\x0442\x044C";
    if (nStringID == STRID_OK)
      return L"\x004F\x004B";
  }
  else
  {
    if (nStringID == STRID_HOTKEY_EXISTS)
      return L"%s\n\nHotkey already assigned to \"%s\".";
    if (nStringID == STRID_CANNOT_OPEN_FILE)
      return L"Cannot open file.";
    if (nStringID == STRID_ALREADY_EXIST)
      return L"Macro already exists.";
    if (nStringID == STRID_CONFIRM_DELETE)
      return L"Are you sure, you want to delete macro?";
    if (nStringID == STRID_THREAD_BUSY)
      return L"Main thread is busy.";
    if (nStringID == STRID_MACRO)
      return L"Macros";
    if (nStringID == STRID_HOTKEY)
      return L"Hotkey";
    if (nStringID == STRID_RECORD_LISTITEM)
      return L"<Record>";
    if (nStringID == STRID_RECORD)
      return L"Record";
    if (nStringID == STRID_STOP)
      return L"Stop";
    if (nStringID == STRID_SAVE)
      return L"Save";
    if (nStringID == STRID_DELETE)
      return L"Delete";
    if (nStringID == STRID_ASSIGN)
      return L"Assign";
    if (nStringID == STRID_OK)
      return L"OK";
  }
  return L"";
}

BOOL IsExtCallParamValid(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return TRUE;
  return FALSE;
}

int GetExtCallParam(LPARAM lParam, int nIndex)
{
  if (*((int *)lParam) >= (int)((nIndex + 1) * sizeof(int)))
    return *(((int *)lParam) + nIndex);
  return 0;
}

void InitCommon(PLUGINDATA *pd)
{
  bInitCommon=TRUE;
  hInstanceDLL=pd->hInstanceDLL;
  hMainWnd=pd->hMainWnd;
  bOldWindows=pd->bOldWindows;
  bMDI=pd->bMDI;
  hMainIcon=pd->hMainIcon;
  wLangSystem=PRIMARYLANGID(pd->wLangSystem);

  if (bOldWindows)
  {
    int i;

    for (i=0; ((char *)pd->pFunction)[i] != ':'; ++i)
      szPluginName[i]=((char *)pd->pFunction)[i];
    szPluginName[i]='\0';
  }
  else
  {
    int i;

    for (i=0; ((wchar_t *)pd->pFunction)[i] != ':'; ++i)
      wszPluginName[i]=((wchar_t *)pd->pFunction)[i];
    wszPluginName[i]='\0';
  }

  if (bOldWindows)
  {
    wsprintfA(szMacrosDir, "%s\\AkelFiles\\Plugs\\Macros", (char *)pd->pAkelDir);
    ReadOptionsA();
  }
  else
  {
    wsprintfW(wszMacrosDir, L"%s\\AkelFiles\\Plugs\\Macros", (wchar_t *)pd->pAkelDir);
    ReadOptionsW();
  }
  dwMainProcessId=GetCurrentProcessId();
  dwMainThreadId=GetCurrentThreadId();
}

void InitMain()
{
  bInitMain=TRUE;

  //SubClass
  NewMainProcData=NULL;
  SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NewMainProc, (LPARAM)&NewMainProcData);
}

void UninitMain()
{
  bInitMain=FALSE;

  if (bOldWindows)
  {
    if (bMainDialogRectSave) SaveOptionsA(OF_WINDOWRECT);
  }
  else
  {
    if (bMainDialogRectSave) SaveOptionsW(OF_WINDOWRECT);
  }

  //Remove subclass
  if (NewMainProcData)
  {
    SendMessage(hMainWnd, AKD_SETMAINPROC, (WPARAM)NULL, (LPARAM)&NewMainProcData);
    NewMainProcData=NULL;
  }

  StackHotkeyFree(&hRecordStack);
  StackHotkeyFree(&hLastMacroStack);
}

//Entry point
extern "C" BOOL WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
  if (fdwReason == DLL_PROCESS_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_ATTACH)
  {
  }
  else if (fdwReason == DLL_THREAD_DETACH)
  {
  }
  else if (fdwReason == DLL_PROCESS_DETACH)
  {
    if (bInitMain) UninitMain();
  }
  return TRUE;
}
